@@ -23,81 +23,74 @@
 /* .fi
 /*	To register a new connection send the following request to
 /*	the \fBanvil\fR(8) server:
-/* .PP
-/* .in +4
-/*	\fBrequest=connect\fR
-/* .br
-/*	\fBident=\fIstring\fR
-/* .in
-/* .PP
+/*
+/* .nf
+/*	    \fBrequest=connect\fR
+/*	    \fBident=\fIstring\fR
+/* .fi
+/*
 /*	The \fBanvil\fR(8) server answers with the number of
 /*	simultaneous connections and the number of connections per
 /*	unit time for the (service, client) combination specified
 /*	with \fBident\fR:
-/* .PP
-/* .in +4
-/*	\fBstatus=0\fR
-/* .br
-/*	\fBcount=\fInumber\fR
-/* .br
-/*	\fBrate=\fInumber\fR
-/* .in
-/* .PP
+/*
+/* .nf
+/*	    \fBstatus=0\fR
+/*	    \fBcount=\fInumber\fR
+/*	    \fBrate=\fInumber\fR
+/* .fi
+/*
 /*	To register a disconnect event send the following request
 /*	to the \fBanvil\fR(8) server:
-/* .PP
-/* .in +4
-/*	\fBrequest=disconnect\fR
-/* .br
-/*	\fBident=\fIstring\fR
-/* .in
-/* .PP
+/*
+/* .nf
+/*	    \fBrequest=disconnect\fR
+/*	    \fBident=\fIstring\fR
+/* .fi
+/*
 /*	The \fBanvil\fR(8) server replies with:
-/* .PP
-/* .ti +4
-/*	\fBstatus=0\fR
+/*
+/* .nf
+/*	    \fBstatus=0\fR
+/* .fi
 /* MESSAGE RATE CONTROL
 /* .ad
 /* .fi
 /*	To register a message delivery request send the following
 /*	request to the \fBanvil\fR(8) server:
-/* .PP
-/* .in +4
-/*	\fBrequest=message\fR
-/* .br
-/*	\fBident=\fIstring\fR
-/* .in
-/* .PP
+/*
+/* .nf
+/*	    \fBrequest=message\fR
+/*	    \fBident=\fIstring\fR
+/* .fi
+/*
 /*	The \fBanvil\fR(8) server answers with the number of message
 /*	delivery requests per unit time for the (service, client)
 /*	combination specified with \fBident\fR:
-/* .PP
-/* .in +4
-/*	\fBstatus=0\fR
-/* .br
-/*	\fBrate=\fInumber\fR
-/* .in
+/*
+/* .nf
+/*	    \fBstatus=0\fR
+/*	    \fBrate=\fInumber\fR
+/* .fi
 /* RECIPIENT RATE CONTROL
 /* .ad
 /* .fi
 /*	To register a recipient request send the following request
 /*	to the \fBanvil\fR(8) server:
-/* .PP
-/* .in +4
-/*	\fBrequest=recipient\fR
-/* .br
-/*	\fBident=\fIstring\fR
-/* .in
-/* .PP
+/*
+/* .nf
+/*	    \fBrequest=recipient\fR
+/*	    \fBident=\fIstring\fR
+/* .fi
+/*
 /*	The \fBanvil\fR(8) server answers with the number of recipient
 /*	addresses per unit time for the (service, client) combination
 /*	specified with \fBident\fR:
-/* .PP
-/* .in +4
-/*	\fBstatus=0\fR
-/* .br
-/*	\fBrate=\fInumber\fR
-/* .in
+/*
+/* .nf
+/*	    \fBstatus=0\fR
+/*	    \fBrate=\fInumber\fR
+/* .fi
 /* TLS SESSION NEGOTIATION RATE CONTROL
 /* .ad
 /* .fi
@@ -106,41 +99,37 @@
 /*
 /*	To register a request for a new (i.e. not cached) TLS session
 /*	send the following request to the \fBanvil\fR(8) server:
-/* .PP
-/* .in +4
-/*	\fBrequest=newtls\fR
-/* .br
-/*	\fBident=\fIstring\fR
-/* .in
-/* .PP
+/*
+/* .nf
+/*	    \fBrequest=newtls\fR
+/*	    \fBident=\fIstring\fR
+/* .fi
+/*
 /*	The \fBanvil\fR(8) server answers with the number of new
 /*	TLS session requests per unit time for the (service, client)
 /*	combination specified with \fBident\fR:
-/* .PP
-/* .in +4
-/*	\fBstatus=0\fR
-/* .br
-/*	\fBrate=\fInumber\fR
-/* .in
-/* .PP
+/*
+/* .nf
+/*	    \fBstatus=0\fR
+/*	    \fBrate=\fInumber\fR
+/* .fi
+/*
 /*	To retrieve new TLS session request rate information without
 /*	updating the counter information, send:
-/* .PP
-/* .in +4
-/*	\fBrequest=newtls_report\fR
-/* .br
-/*	\fBident=\fIstring\fR
-/* .in
-/* .PP
+/*
+/* .nf
+/*	    \fBrequest=newtls_report\fR
+/*	    \fBident=\fIstring\fR
+/* .fi
+/*
 /*	The \fBanvil\fR(8) server answers with the number of new
 /*	TLS session requests per unit time for the (service, client)
 /*	combination specified with \fBident\fR:
-/* .PP
-/* .in +4
-/*	\fBstatus=0\fR
-/* .br
-/*	\fBrate=\fInumber\fR
-/* .in
+/*
+/* .nf
+/*	    \fBstatus=0\fR
+/*	    \fBrate=\fInumber\fR
+/* .fi
 /* SECURITY
 /* .ad
 /* .fi
@@ -173,8 +162,8 @@
 /*
 /*	In this preliminary implementation, a count (or rate) limited server
 /*	can have only one remote client at a time. If a server reports
-/*	multiple simultaneous clients, all but the last reported client
-/*	are ignored.
+/*	multiple simultaneous clients, state is kept only for the last
+/*	reported client.
 /*
 /*	The \fBanvil\fR(8) server automatically discards client
 /*	request information after it expires.  To prevent the
@@ -208,11 +197,11 @@
 /*	The time limit for sending or receiving information over an internal
 /*	communication channel.
 /* .IP "\fBmax_idle (100s)\fR"
-/*	The maximum amount of time that an idle Postfix daemon process
-/*	waits for the next service request before exiting.
+/*	The maximum amount of time that an idle Postfix daemon process waits
+/*	for an incoming connection before terminating voluntarily.
 /* .IP "\fBmax_use (100)\fR"
-/*	The maximal number of connection requests before a Postfix daemon
-/*	process terminates.
+/*	The maximal number of incoming connections that a Postfix daemon
+/*	process will service before terminating voluntarily.
 /* .IP "\fBprocess_id (read-only)\fR"
 /*	The process ID of a Postfix command or daemon process.
 /* .IP "\fBprocess_name (read-only)\fR"
@@ -267,6 +256,7 @@
 
 #include <mail_conf.h>
 #include <mail_params.h>
+#include <mail_version.h>
 #include <mail_proto.h>
 #include <anvil_clnt.h>
 
@@ -498,7 +488,7 @@ static time_t max_cache_time;		/* time of peak size */
 static void anvil_remote_expire(int unused_event, char *context)
 {
     ANVIL_REMOTE *anvil_remote = (ANVIL_REMOTE *) context;
-    char   *myname = "anvil_remote_expire";
+    const char *myname = "anvil_remote_expire";
 
     if (msg_verbose)
 	msg_info("%s %s", myname, anvil_remote->ident);
@@ -521,7 +511,7 @@ static void anvil_remote_expire(int unused_event, char *context)
 static void anvil_remote_lookup(VSTREAM *client_stream, const char *ident)
 {
     ANVIL_REMOTE *anvil_remote;
-    char   *myname = "anvil_remote_lookup";
+    const char *myname = "anvil_remote_lookup";
 
     if (msg_verbose)
 	msg_info("%s fd=%d stream=0x%lx ident=%s",
@@ -566,7 +556,7 @@ static ANVIL_REMOTE *anvil_remote_conn_update(VSTREAM *client_stream, const char
 {
     ANVIL_REMOTE *anvil_remote;
     ANVIL_LOCAL *anvil_local;
-    char   *myname = "anvil_remote_conn_update";
+    const char *myname = "anvil_remote_conn_update";
 
     if (msg_verbose)
 	msg_info("%s fd=%d stream=0x%lx ident=%s",
@@ -767,7 +757,7 @@ static void anvil_remote_disconnect(VSTREAM *client_stream, const char *ident)
 {
     ANVIL_REMOTE *anvil_remote;
     ANVIL_LOCAL *anvil_local;
-    char   *myname = "anvil_remote_disconnect";
+    const char *myname = "anvil_remote_disconnect";
 
     if (msg_verbose)
 	msg_info("%s fd=%d stream=0x%lx ident=%s",
@@ -803,7 +793,7 @@ static void anvil_service_done(VSTREAM *client_stream, char *unused_service,
 			               char **unused_argv)
 {
     ANVIL_LOCAL *anvil_local;
-    char   *myname = "anvil_service_done";
+    const char *myname = "anvil_service_done";
 
     if (msg_verbose)
 	msg_info("%s fd=%d stream=0x%lx",
@@ -945,6 +935,8 @@ static void post_jail_init(char *unused_name, char **unused_argv)
 	var_idle_limit = var_anvil_time_unit;
 }
 
+MAIL_VERSION_STAMP_DECLARE;
+
 /* main - pass control to the multi-threaded skeleton */
 
 int     main(int argc, char **argv)
@@ -955,6 +947,11 @@ int     main(int argc, char **argv)
 	0,
     };
 
+    /*
+     * Fingerprint executables and core dumps.
+     */
+    MAIL_VERSION_STAMP_ALLOCATE;
+
     multi_server_main(argc, argv, anvil_service,
 		      MAIL_SERVER_TIME_TABLE, time_table,
 		      MAIL_SERVER_POST_INIT, post_jail_init,
