@@ -31,6 +31,12 @@
 #include <mail_conf.h>
 #include <mime_state.h>
 #include <string_list.h>
+#include <cleanup_user.h>
+
+ /*
+  * Milter library.
+  */
+#include <milter.h>
 
  /*
   * These state variables are accessed by many functions, and there is only
@@ -41,6 +47,7 @@ typedef struct CLEANUP_STATE {
     VSTRING *temp1;			/* scratch buffer, local use only */
     VSTRING *temp2;			/* scratch buffer, local use only */
     VSTRING *stripped_buf;		/* character stripped input */
+    VSTREAM *src;			/* current input stream */
     VSTREAM *dst;			/* current output stream */
     MAIL_STREAM *handle;		/* mail stream handle */
     char   *queue_name;			/* queue name */
@@ -62,7 +69,13 @@ typedef struct CLEANUP_STATE {
     BH_TABLE *dups;			/* recipient dup filter */
     void    (*action) (struct CLEANUP_STATE *, int, const char *, ssize_t);
     off_t   data_offset;		/* start of message content */
+    off_t   body_offset;		/* start of body content */
     off_t   xtra_offset;		/* start of extra segment */
+    off_t   cont_length;		/* length including Milter edits */
+    off_t   append_rcpt_pt_offset;	/* append recipient here */
+    off_t   append_rcpt_pt_target;	/* target of above record */
+    off_t   append_hdr_pt_offset;	/* append header here */
+    off_t   append_hdr_pt_target;	/* target of above record */
     ssize_t rcpt_count;			/* recipient count */
     char   *reason;			/* failure reason */
     NVTABLE *attr;			/* queue file attribute list */
@@ -79,6 +92,25 @@ typedef struct CLEANUP_STATE {
 #ifdef DELAY_ACTION
     int     defer_delay;		/* deferred delivery */
 #endif
+
+    /*
+     * Miscellaneous Milter support.
+     */
+    MILTERS *milters;			/* mail filters */
+    const char *client_name;		/* real or ersatz client */
+    const char *reverse_name;		/* real or ersatz client */
+    const char *client_addr;		/* real or ersatz client */
+    int     client_af;			/* real or ersatz client */
+    const char *client_port;		/* real or ersatz client */
+    VSTRING *milter_ext_from;		/* externalized sender */
+    VSTRING *milter_ext_rcpt;		/* externalized recipient */
+
+    /*
+     * Support for Milter body replacement requests.
+     */
+    struct CLEANUP_REGION *free_regions;/* unused regions */
+    struct CLEANUP_REGION *body_regions;/* regions with body content */
+    struct CLEANUP_REGION *curr_body_region;
 } CLEANUP_STATE;
 
  /*
@@ -114,6 +146,11 @@ extern MAPS *cleanup_rcpt_bcc_maps;
 extern VSTRING *cleanup_reject_chars;
 extern VSTRING *cleanup_strip_chars;
 
+ /*
+  * Milters.
+  */
+extern MILTERS *cleanup_milters;
+
  /*
   * Address canonicalization fine control.
   */
@@ -146,13 +183,13 @@ extern VSTRING *cleanup_bounce_path;
  /*
   * cleanup_state.c
   */
-extern CLEANUP_STATE *cleanup_state_alloc(void);
+extern CLEANUP_STATE *cleanup_state_alloc(VSTREAM *);
 extern void cleanup_state_free(CLEANUP_STATE *);
 
  /*
   * cleanup_api.c
   */
-extern CLEANUP_STATE *cleanup_open(void);
+extern CLEANUP_STATE *cleanup_open(VSTREAM *);
 extern void cleanup_control(CLEANUP_STATE *, int);
 extern int cleanup_flush(CLEANUP_STATE *);
 extern void cleanup_free(CLEANUP_STATE *);
@@ -173,6 +210,7 @@ extern CONFIG_TIME_TABLE cleanup_time_table[];
 extern void cleanup_out(CLEANUP_STATE *, int, const char *, ssize_t);
 extern void cleanup_out_string(CLEANUP_STATE *, int, const char *);
 extern void PRINTFLIKE(3, 4) cleanup_out_format(CLEANUP_STATE *, int, const char *,...);
+extern void cleanup_out_header(CLEANUP_STATE *, VSTRING *);
 
 #define CLEANUP_OUT_BUF(s, t, b) \
 	cleanup_out((s), (t), vstring_str((b)), VSTRING_LEN((b)))
@@ -195,6 +233,11 @@ extern void cleanup_message(CLEANUP_STATE *, int, const char *, ssize_t);
   */
 extern void cleanup_extracted(CLEANUP_STATE *, int, const char *, ssize_t);
 
+ /*
+  * cleanup_final.c
+  */
+extern void cleanup_final(CLEANUP_STATE *);
+
  /*
   * cleanup_rewrite.c
   */
@@ -244,6 +287,40 @@ extern int cleanup_bounce(CLEANUP_STATE *);
 #define CLEANUP_MSG_STATS(stats, state) \
     MSG_STATS_INIT1(stats, incoming_arrival, state->arrival_time)
 
+ /*
+  * cleanup_milter.c.
+  */
+extern void cleanup_milter_receive(CLEANUP_STATE *, int);
+extern void cleanup_milter_inspect(CLEANUP_STATE *, MILTERS *);
+extern void cleanup_milter_emul_mail(CLEANUP_STATE *, MILTERS *, const char *);
+extern void cleanup_milter_emul_rcpt(CLEANUP_STATE *, MILTERS *, const char *);
+extern void cleanup_milter_emul_data(CLEANUP_STATE *, MILTERS *);
+
+#define CLEANUP_MILTER_OK(s) \
+    (((s)->flags & CLEANUP_FLAG_MILTER) != 0 \
+	&& (s)->errs == 0 && ((s)->flags & CLEANUP_FLAG_DISCARD) == 0)
+
+ /*
+  * cleanup_body_edit.c
+  */
+typedef struct CLEANUP_REGION {
+    off_t   start;			/* start of region */
+    off_t   len;			/* length or zero (open-ended) */
+    off_t   write_offs;			/* write offset */
+    struct CLEANUP_REGION *next;	/* linkage */
+} CLEANUP_REGION;
+
+extern void cleanup_region_init(CLEANUP_STATE *);
+extern CLEANUP_REGION *cleanup_region_open(CLEANUP_STATE *, ssize_t);
+extern void cleanup_region_close(CLEANUP_STATE *, CLEANUP_REGION *);
+extern CLEANUP_REGION *cleanup_region_return(CLEANUP_STATE *, CLEANUP_REGION *);
+extern void cleanup_region_done(CLEANUP_STATE *);
+
+extern int cleanup_body_edit_start(CLEANUP_STATE *);
+extern int cleanup_body_edit_write(CLEANUP_STATE *, int, VSTRING *);
+extern int cleanup_body_edit_finish(CLEANUP_STATE *);
+extern void cleanup_body_edit_free(CLEANUP_STATE *);
+
 /* LICENSE
 /* .ad
 /* .fi
