@@ -63,7 +63,6 @@
 #include <stdlib.h>
 #include <string.h>
 #include <stdarg.h>
-#include <limits.h>			/* INT_MAX */
 
 #ifndef SHUT_RDWR
 #define SHUT_RDWR	2
@@ -201,6 +200,7 @@
 
 /* Global library. */
 
+#include <mail_params.h>		/* var_line_limit */
 #include <mail_proto.h>
 #include <rec_type.h>
 #include <record.h>
@@ -281,7 +281,7 @@ typedef struct {
  /*
   * We don't accept insane amounts of data.
   */
-#define XXX_MAX_DATA	(INT_MAX / 2)
+#define XXX_MAX_DATA	(MILTER_CHUNK_SIZE * 2)
 #define XXX_TIMEOUT	10
 
 #ifndef USE_LIBMILTER_INCLUDES
@@ -889,11 +889,13 @@ static const char *milter8_event(MILTER8 *milter, int event,
     ssize_t data_size;
     const char *smfic_name;
     const char *smfir_name;
-    MILTERS *parent;
+    MILTERS *parent = milter->m.parent;
     UINT32_TYPE index;
     const char *edit_resp = 0;
     const char *retval = 0;
+    VSTRING *body_line_buf = 0;
     int     done = 0;
+    int     body_edit_lockout = 0;
 
 #define DONT_SKIP_REPLY	0
 
@@ -995,7 +997,7 @@ static const char *milter8_event(MILTER8 *milter, int event,
     /*
      * Receive the reply or replies.
      * 
-     * Intercept all loop exits so that we can do post-(queue file edit)
+     * Intercept all loop exits so that we can do post header/body edit
      * processing.
      * 
      * XXX Bound the loop iteration count.
@@ -1007,7 +1009,9 @@ static const char *milter8_event(MILTER8 *milter, int event,
      * Instead skip all further Milter replies until the final decision. This
      * way the Postfix SMTP server stays in sync with the Milter, and Postfix
      * doesn't have to lose the ability to handle multiple deliveries within
-     * the same SMTP session.
+     * the same SMTP session. This requires that the Postfix SMTP server uses
+     * something other than CLEANUP_STAT_WRITE when it loses contact with the
+     * cleanup server.
      */
 #define IN_CONNECT_EVENT(e) ((e) == SMFIC_CONNECT || (e) == SMFIC_HELO)
 
@@ -1032,6 +1036,32 @@ static const char *milter8_event(MILTER8 *milter, int event,
 	    msg_info("reply: %s data %ld bytes",
 		     (smfir_name = str_name_code(smfir_table, cmd)) != 0 ?
 		     smfir_name : "unknown", (long) data_size);
+
+	/*
+	 * Handle unfinished message body replacement first.
+	 * 
+	 * XXX When SMFIR_REPLBODY is followed by some different request, we
+	 * assume that the body replacement operation is complete. The queue
+	 * file editing implementation currently does not support sending
+	 * part 1 of the body replacement text, doing some other queue file
+	 * updates, and then sending part 2 of the body replacement text. To
+	 * avoid loss of data, we log an error when SMFIR_REPLBODY requests
+	 * are alternated with other requests.
+	 */
+	if (body_line_buf != 0 && cmd != SMFIR_REPLBODY) {
+	    /* In case the last body replacement line didn't end in CRLF. */
+	    if (edit_resp == 0 && LEN(body_line_buf) > 0)
+		edit_resp = parent->repl_body(parent->chg_context,
+					      MILTER_BODY_LINE,
+					      body_line_buf);
+	    if (edit_resp == 0)
+		edit_resp = parent->repl_body(parent->chg_context,
+					      MILTER_BODY_END,
+					      (VSTRING *) 0);
+	    body_edit_lockout = 1;
+	    vstring_free(body_line_buf);
+	    body_line_buf = 0;
+	}
 	switch (cmd) {
 
 	    /*
@@ -1186,8 +1216,7 @@ static const char *milter8_event(MILTER8 *milter, int event,
 	    /*
 	     * Decision: quarantine. In Sendmail 8.13 this does not imply a
 	     * transition in the receiver state (reply, reject, tempfail,
-	     * accept, discard). We should not transition, either, otherwise
-	     * we get out of sync.
+	     * accept, discard).
 	     */
 #ifdef SMFIR_QUARANTINE
 	case SMFIR_QUARANTINE:
@@ -1196,8 +1225,7 @@ static const char *milter8_event(MILTER8 *milter, int event,
 				  MILTER8_DATA_BUFFER, milter->buf,
 				  MILTER8_DATA_END) != 0)
 		MILTER8_EVENT_BREAK(milter->def_reply);
-	    milter8_def_reply(milter, "H");
-	    continue;
+	    MILTER8_EVENT_BREAK("H");
 #endif
 
 	    /*
@@ -1222,7 +1250,6 @@ static const char *milter8_event(MILTER8 *milter, int event,
 		    /* Skip to the next request after previous edit error. */
 		    if (edit_resp)
 			continue;
-		    parent = milter->m.parent;
 		    /* XXX Sendmail 8 compatibility. */
 		    if (index == 0)
 			index = 1;
@@ -1262,7 +1289,6 @@ static const char *milter8_event(MILTER8 *milter, int event,
 		    /* Skip to the next request after previous edit error. */
 		    if (edit_resp)
 			continue;
-		    parent = milter->m.parent;
 		    edit_resp = parent->add_header(parent->chg_context,
 						   STR(milter->buf),
 						   STR(milter->body));
@@ -1291,7 +1317,6 @@ static const char *milter8_event(MILTER8 *milter, int event,
 			milter8_conf_error(milter);
 			MILTER8_EVENT_BREAK(milter->def_reply);
 		    }
-		    parent = milter->m.parent;
 		    edit_resp = parent->ins_header(parent->chg_context,
 						   (ssize_t) index + 1,
 						   STR(milter->buf),
@@ -1310,7 +1335,6 @@ static const char *milter8_event(MILTER8 *milter, int event,
 		    /* Skip to the next request after previous edit error. */
 		    if (edit_resp)
 			continue;
-		    parent = milter->m.parent;
 		    edit_resp = parent->add_rcpt(parent->chg_context,
 						 STR(milter->buf));
 		    continue;
@@ -1326,7 +1350,6 @@ static const char *milter8_event(MILTER8 *milter, int event,
 		    /* Skip to the next request after previous edit error. */
 		    if (edit_resp)
 			continue;
-		    parent = milter->m.parent;
 		    edit_resp = parent->del_rcpt(parent->chg_context,
 						 STR(milter->buf));
 		    continue;
@@ -1335,8 +1358,14 @@ static const char *milter8_event(MILTER8 *milter, int event,
 		     * Modification request: replace the message body, and
 		     * update the message size.
 		     */
-#if 0
 		case SMFIR_REPLBODY:
+		    if (body_edit_lockout) {
+			msg_warn("milter %s: body replacement requests can't "
+				 "currently be mixed with other requests",
+				 milter->m.name);
+			milter8_conf_error(milter);
+			MILTER8_EVENT_BREAK(milter->def_reply);
+		    }
 		    if (milter8_read_data(milter, data_size,
 					  MILTER8_DATA_BUFFER, milter->body,
 					  MILTER8_DATA_END) != 0)
@@ -1344,11 +1373,31 @@ static const char *milter8_event(MILTER8 *milter, int event,
 		    /* Skip to the next request after previous edit error. */
 		    if (edit_resp)
 			continue;
-		    parent = milter->m.parent;
-		    edit_resp = parent->repl_body(parent->chg_context,
-						  milter->body);
+		    /* Start body replacement. */
+		    if (body_line_buf == 0) {
+			body_line_buf = vstring_alloc(var_line_limit);
+			edit_resp = parent->repl_body(parent->chg_context,
+						      MILTER_BODY_START,
+						      (VSTRING *) 0);
+		    }
+		    /* Extract lines from the on-the-wire CRLF format. */
+		    for (cp = STR(milter->body); edit_resp == 0
+			 && cp < vstring_end(milter->body); cp++) {
+			ch = *(unsigned char *) cp;
+			if (ch == '\n') {
+			    if (LEN(body_line_buf) > 0
+				&& vstring_end(body_line_buf)[-1] == '\r')
+				vstring_truncate(body_line_buf,
+						 LEN(body_line_buf) - 1);
+			    edit_resp = parent->repl_body(parent->chg_context,
+							  MILTER_BODY_LINE,
+							  body_line_buf);
+			    VSTRING_RESET(body_line_buf);
+			} else {
+			    VSTRING_ADDCH(body_line_buf, ch);
+			}
+		    }
 		    continue;
-#endif
 		}
 	    }
 	    msg_warn("milter %s: unexpected filter response %s after event %s",
@@ -1372,6 +1421,12 @@ static const char *milter8_event(MILTER8 *milter, int event,
 	MILTER8_EVENT_BREAK(milter->def_reply);
     }
 
+    /*
+     * Clean up after aborted message body replacement.
+     */
+    if (body_line_buf)
+	vstring_free(body_line_buf);
+
     /*
      * XXX Some cleanup clients ask the cleanup server to bounce mail for
      * them. In that case we must override a hard reject retval result after
@@ -1397,7 +1452,7 @@ static void milter8_connect(MILTER8 *milter)
     int     fd;
     const UINT32_TYPE my_actions = (SMFIF_ADDHDRS | SMFIF_ADDRCPT
 				    | SMFIF_DELRCPT | SMFIF_CHGHDRS
-    /* Not yet: | SMFIF_CHGBODY */
+				    | SMFIF_CHGBODY
 #ifdef SMFIF_QUARANTINE
 				    | SMFIF_QUARANTINE
 #endif
@@ -1627,17 +1682,6 @@ static const char *milter8_conn_event(MILTER *m,
     const char *myname = "milter8_conn_event";
     MILTER8 *milter = (MILTER8 *) m;
     int     port;
-    const char *sm_name;
-    char   *ptr = 0;
-    const char *resp;
-
-    /*
-     * Need a global definition for "unknown" host name or address that is
-     * shared by smtpd, cleanup and libmilter.
-     */
-#define XXX_UNKNOWN	"unknown"
-#define STR_EQ(x,y)	(strcmp((x), (y)) == 0)
-#define STR_NE(x,y)	(strcmp((x), (y)) != 0)
 
     /*
      * XXX Sendmail 8 libmilter closes the MTA-to-filter socket when it finds
@@ -1669,51 +1713,40 @@ static const char *milter8_conn_event(MILTER *m,
 	    port = 0;
 	}
 	milter->state = MILTER8_STAT_ENVELOPE;
-	/* Transform unknown hostname from Postfix to Sendmail form. */
-	sm_name = (STR_NE(client_name, XXX_UNKNOWN) ? client_name :
-		   STR_EQ(client_addr, XXX_UNKNOWN) ? client_name :
-		   (ptr = concatenate("[", client_addr, "]", (char *) 0)));
 	switch (addr_family) {
 	case AF_INET:
-	    resp = milter8_event(milter, SMFIC_CONNECT, SMFIP_NOCONNECT,
-				 DONT_SKIP_REPLY, macros,
-				 MILTER8_DATA_STRING, sm_name,
-				 MILTER8_DATA_OCTET, SMFIA_INET,
-				 MILTER8_DATA_NSHORT, htons(port),
-				 MILTER8_DATA_STRING, client_addr,
-				 MILTER8_DATA_END);
-	    break;
+	    return (milter8_event(milter, SMFIC_CONNECT, SMFIP_NOCONNECT,
+				  DONT_SKIP_REPLY, macros,
+				  MILTER8_DATA_STRING, client_name,
+				  MILTER8_DATA_OCTET, SMFIA_INET,
+				  MILTER8_DATA_NSHORT, htons(port),
+				  MILTER8_DATA_STRING, client_addr,
+				  MILTER8_DATA_END));
 #ifdef HAS_IPV6
 	case AF_INET6:
-	    resp = milter8_event(milter, SMFIC_CONNECT, SMFIP_NOCONNECT,
-				 DONT_SKIP_REPLY, macros,
-				 MILTER8_DATA_STRING, sm_name,
-				 MILTER8_DATA_OCTET, SMFIA_INET6,
-				 MILTER8_DATA_NSHORT, htons(port),
-				 MILTER8_DATA_STRING, client_addr,
-				 MILTER8_DATA_END);
-	    break;
+	    return (milter8_event(milter, SMFIC_CONNECT, SMFIP_NOCONNECT,
+				  DONT_SKIP_REPLY, macros,
+				  MILTER8_DATA_STRING, client_name,
+				  MILTER8_DATA_OCTET, SMFIA_INET6,
+				  MILTER8_DATA_NSHORT, htons(port),
+				  MILTER8_DATA_STRING, client_addr,
+				  MILTER8_DATA_END));
 #endif
 	case AF_UNIX:
-	    resp = milter8_event(milter, SMFIC_CONNECT, SMFIP_NOCONNECT,
-				 DONT_SKIP_REPLY, macros,
-				 MILTER8_DATA_STRING, sm_name,
-				 MILTER8_DATA_OCTET, SMFIA_UNIX,
-				 MILTER8_DATA_NSHORT, htons(0),
-				 MILTER8_DATA_STRING, client_addr,
-				 MILTER8_DATA_END);
-	    break;
+	    return (milter8_event(milter, SMFIC_CONNECT, SMFIP_NOCONNECT,
+				  DONT_SKIP_REPLY, macros,
+				  MILTER8_DATA_STRING, client_name,
+				  MILTER8_DATA_OCTET, SMFIA_UNIX,
+				  MILTER8_DATA_NSHORT, htons(0),
+				  MILTER8_DATA_STRING, client_addr,
+				  MILTER8_DATA_END));
 	default:
-	    resp = milter8_event(milter, SMFIC_CONNECT, SMFIP_NOCONNECT,
-				 DONT_SKIP_REPLY, macros,
-				 MILTER8_DATA_STRING, sm_name,
-				 MILTER8_DATA_OCTET, SMFIA_UNKNOWN,
-				 MILTER8_DATA_END);
-	    break;
+	    return (milter8_event(milter, SMFIC_CONNECT, SMFIP_NOCONNECT,
+				  DONT_SKIP_REPLY, macros,
+				  MILTER8_DATA_STRING, client_name,
+				  MILTER8_DATA_OCTET, SMFIA_UNKNOWN,
+				  MILTER8_DATA_END));
 	}
-	if (ptr != 0)
-	    myfree(ptr);
-	return (resp);
     default:
 	msg_panic("%s: milter %s: bad state %d",
 		  myname, milter->m.name, milter->state);
@@ -1741,6 +1774,7 @@ static const char *milter8_helo_event(MILTER *m, const char *helo_name,
 	return (milter->def_reply);
     case MILTER8_STAT_ENVELOPE:
     case MILTER8_STAT_ACCEPT_MSG:
+	/* With HELO after MAIL, smtpd(8) calls milter8_abort() next. */
 	if (msg_verbose)
 	    msg_info("%s: milter %s: helo %s",
 		     myname, milter->m.name, helo_name);
