@@ -138,7 +138,7 @@ static void resolve_addr(RES_CONTEXT *rp, char *sender, char *addr,
 			         VSTRING *channel, VSTRING *nexthop,
 			         VSTRING *nextrcpt, int *flags)
 {
-    char   *myname = "resolve_addr";
+    const char *myname = "resolve_addr";
     VSTRING *addr_buf = vstring_alloc(100);
     TOK822 *tree = 0;
     TOK822 *saved_domain = 0;
@@ -620,7 +620,8 @@ static void resolve_addr(RES_CONTEXT *rp, char *sender, char *addr,
 	if ((newloc = mail_addr_find(relocated_maps, STR(nextrcpt),
 				     IGNORE_ADDR_EXTENSION)) != 0) {
 	    vstring_strcpy(channel, MAIL_SERVICE_ERROR);
-	    vstring_sprintf(nexthop, "User has moved to %s", newloc);
+	    /* 5.1.6 is the closest match, but not perfect. */
+	    vstring_sprintf(nexthop, "5.1.6 User has moved to %s", newloc);
 	} else if (dict_errno != 0) {
 	    msg_warn("%s lookup failure", VAR_RELOCATED_MAPS);
 	    *flags |= RESOLVE_FLAG_FAIL;
